class Context:
    def __init__(self, socket, data):
        injected = data.get('injected', None)
        self.socket = socket
        self.command = data['data'].get('name', None)
        self.command_id = data['data'].get('id', None)
        self.ucid = data['data'].get('custom_id')
        self.channel_id = data.get('channel_id', None)
        self.id = data['id']
        self.token = data['token']
        self.callback_url = f"https://discord.com/api/v8/interactions/{self.id}/{self.token}/callback"
        if injected is not None:
            self.type = injected['type']

    async def callback(self, content: str = '', embeds: list = [], components:list = [], mentions: list = []):
        cleaned_components = []
        for ar in components:
            new_ar = {"type": 1, "components": []}
            for component in ar['components']:
                if isinstance(component, dict):
                    new_ar['components'].append(component)
                else:
                    new_ar['components'].append(component.build())
            cleaned_components.append(new_ar)

        message = {
            "type": 4,
            "data": {
                "content": content,
                "embeds": embeds,
                "components": cleaned_components,
                "allowed_mentions": mentions
            }
        }

        async with self.socket.session.post(self.callback_url, json=message, headers=self.socket.headers) as maybe_send:
            if maybe_send.status not in [200, 201, 202, 204]:
                print(await maybe_send.json())




import inspect


class Event:
    def __init__(self, name, func):
        self.name = name
        self.func = func

def event(name):
    def predicate(func):
        if not inspect.iscoroutinefunction(func):
            raise TypeError("Event functions must be coroutine.")
        return Event(name, func)
    return predicate


from enum import Enum

class Avatars(Enum):
    blurlpe = 0
    grey = 1
    gray = 1
    green = 2
    orange = 3
    red = 4

    def __str__(self):
        return self.name

def return_cdn_avatar(data):
    if data['avatar'] is not None:
        animated = data.startswith('a_')
        form = 'gif' if animated else 'png'
        return f"https://cdn.discordapp.com/avatars/{data['id']}/{data['avatar']}.{form}?size=2024"
    else:
        return f"https://cdn.discordapp.com/embed/avatars/{int(data['discriminator']) % len(Avatars)}.png"


from .. import utils

class User:
    """Represents a Discord user"""
    def __init__(self, data):
        self.data = data
        self.id = int(data['id']) # Discord sends the id as a string, convert to int
        self.discriminator = int(data['discriminator'])
        self.name = data['username']

    @property
    def avatar(self):
        return utils.return_cdn_avatar(self.data)
    
    @property
    def username(self):
        return str(self.name) + "#" + str(self.discriminator)



class ButtonStyle:
    PRIMARY = 1
    SECONDARY = 2
    SUCCESS = 3
    DANGER = 4
    LINK = 5

class Button:
    def __init__(
        self,
        style: int = ButtonStyle.PRIMARY,
        label: str = '',
        custom_id: str = '',
        url: str = None,
        emoji: dict = None,
        disabled: bool = False
    ):  
        self.base = {
            "type": 2,
            "style": style,
            "label": label,
            "custom_id": custom_id,
            "disabled": disabled
        }

        if emoji is not None:
            self.base['emoji'] = emoji
        
        if style == ButtonStyle.LINK:
            self.base["url"] = url

    def build(self):
        return self.base


from typing import Optional
import aiohttp
import asyncio
import json
import importlib
import threading
import time
import concurrent.futures
import sys
import traceback
import inspect

from . import utils
from .listener import EventContainer
from .commands import Command
from .components import Component
from .events import Event
from .models.context import Context
from .models.user import User


class MaintainSocketAlive(threading.Thread):
    def __init__(self, *args, **kwargs):
        socket = kwargs.pop('socket')
        interval = kwargs.pop('interval')
        threading.Thread.__init__(self, *args, **kwargs)
        self.socket = socket
        self.interval = interval / 1000
        self.daemon = True
        self._stop_event = threading.Event()
        self._last_ack = time.perf_counter()
        self._last_recv = time.perf_counter()
        self._last_send = time.perf_counter()
        self.latency = float('inf')
        self.main_id = socket.thread_id

    def run(self):
        while not self._stop_event.wait(self.interval):
            if self._last_recv + 60 < time.perf_counter():
                func = self.socket.close(4000)
                f = asyncio.run_coroutine_threadsafe(func, loop=self.socket.loop)
                try:
                    f.result()
                except Exception:
                    pass
                finally:
                    self.stop()
                    return
            
            coro = self.socket.send_heartbeat(self.payload())
            f = asyncio.run_coroutine_threadsafe(coro, loop=self.socket.loop)
            try:
                total = 0
                while True:
                    try:
                        f.result(10)
                        break
                    except concurrent.futures.TimeoutError:
                        total += 10
                        try:
                            frame = sys._current_frames()[self.main_id]
                        except KeyError:
                            pass
            
            except Exception:
                self.stop()
                traceback.print_exc()
            else:
                self._last_send = time.perf_counter()

    def payload(self):
        return {
            "op": 1,
            "d": None
        }
    def stop(self):
        self._stop_event.set()

    def tick(self):
        self._last_recv = time.perf_counter()
    def ack(self):
        ack_time = time.perf_counter()
        self._last_ack = ack_time
        self.latency = ack_time - self._last_send


class Socket:
    """Connects to the Discord websocket"""
    def __init__(self, loop: Optional[asyncio.AbstractEventLoop] = None):
        self.loop: asyncio.AbstractEventLoop = asyncio.get_event_loop() if loop is None else loop
        self.session: aiohttp.BaseConnector = None
        self.__token: str = None
        self.headers: dict = None
        self.user = None
        self.alive_handler = None
        self.events = EventContainer()
        self.thread_id = threading.get_ident()
        self.connection = None
        self.unchecked_decorators = {"event": [], "command": [], "component": []}

    def add_component_listener(self, ucid, func):
        if not inspect.iscoroutinefunction(func):
            raise TypeError("Component functions must be a coroutine")
        self.events.component(Component(ucid, func))

    def event(self, name):
        def predicate(func):
            if not inspect.iscoroutinefunction(func):
                raise TypeError("Event function must be coroutine")
            self.unchecked_decorators['event'].append(Event(name, func))
        return predicate
    
    def command(self, name, _type):
        def predicate(func):
            if not inspect.iscoroutinefunction(func):
                raise TypeError("Command functions must be coroutine")
            self.unchecked_decorators['command'].append(Command(name, func, _type))
        return predicate

    def component(self, ucid):
        def predicate(func):
            if not inspect.iscoroutinefunction(func):
                raise TypeError("Component function must be coroutine")
            self.unchecked_decorators['component'].append(Component(ucid, func))
        return predicate

    async def send_heartbeat(self, payload):
        await self.connection.send_json(payload)

    async def run_command(self, context):
        coro = None
        if context.type == utils.SLASH:
            coro = self.events.commands.get(context.command, None)
        elif context.type == utils.USER:
            coro = self.events.user_commands.get(context.command, None)
        elif context.type == utils.MESSAGE:
            coro = self.events.message_commands.get(context.command, None)

        if coro is None:
            raise ValueError(f"No command with name {context.name} was found.")
        
        await coro(context)

    async def run_component_command(self, context):
        coro = None
        coro = self.events.components.get(context.ucid, None)
        
        if coro is None:
            raise ValueError(f"No component with unique id {context.ucid} was found.")
        try:
            print(coro)
            await coro(context)
        except Exception:
            traceback.print_exc()

    async def run_ready_event(self):
        coro = None
        coro = self.events.events['on_ready'][0]
        if coro is not None:
            await coro()

    async def before_login(self):
        # Construct headers
        self.headers = {"Authorization": f"Bot {self.__token}"}
        self.session = aiohttp.ClientSession()
        # Make user
        self.user = User(await (await self.session.get('https://discord.com/api/v8/users/@me', headers=self.headers)).json())

    async def connect(self):
        await self.before_login()
        async with self.session.ws_connect("wss://gateway.discord.gg/") as connection:
            self.connection = connection
            async for message in connection:
                msg = json.loads(message.data)
                op, t , d = msg['op'], msg['t'], msg['d']

                if self.alive_handler:
                    self.alive_handler.tick()

                try:    

                    if op == utils.HELLO: # Discord's 'Hello' payload
                        await connection.send_json(
                            {
                                "op": utils.IDENTIFIY,
                                "d": {
                                    "token": self.__token,
                                    "intents": 513,
                                    "properties": {
                                        "$os": "Windows",
                                        "$browser": "discsocket 0.0.1",
                                        "$device": "discsocket 0.0.1"
                                    },
                                    "compress": False,
                                    "large_threshold": 250
                                }
                            }   
                        )
                        self.alive_handler = MaintainSocketAlive(socket=self, interval=d['heartbeat_interval'])
                        await connection.send_json(self.alive_handler.payload())
                        self.alive_handler.start()
                
                    elif op == utils.HEARTBEAT_ACK:
                        if self.alive_handler:
                            self.alive_handler.ack()
                
                    elif op ==utils.HEARTBEAT:
                        if self.alive_handler:
                            connection.send_json(self.alive_handler.payload())

                    elif t == 'INTERACTION_CREATE':
                        if d['type'] == 2: # Slash command
                            # inject type
                            d['injected'] = {"type": utils.SLASH}
                            model = Context(self, d)
                            await self.run_command(model)
                        elif d['type'] == 3: # Component interaction
                            model = Context(self, d)
                            await self.run_component_command(model)
                    elif t == 'READY':
                        await self.run_ready_event()
                except Exception:
                    traceback.print_exc()

    def add_extension(self, extension_path):
        extension = importlib.import_module(extension_path)
        for name in dir(extension):
            attr = getattr(extension, name)
            
            if isinstance(attr, Command):
                self.events.command(attr)
            elif isinstance(attr, Component):
                self.events.component(attr)
            elif isinstance(attr, Event):
                self.events.event(attr)
                    
    def run(self, token: str):

        for _type in ['event', 'command', 'component']:
            if len(self.unchecked_decorators[_type]) > 0:
                for listener in self.unchecked_decorators[_type]:
                    if isinstance(listener, Command):
                        self.events.command(listener)
                    elif isinstance(listener, Event):
                        self.events.event(listener)
                    elif isinstance(listener, Component):
                        self.events.component(listener)

        self.__token = token
        self.loop.create_task(self.connect())
        self.loop.run_forever()
    


from setuptools import setup

readme = ''
with open('README.md') as f:
    readme = f.read()


setup(
    name="discsocket",
    author="Tadeo Murillo",
    url="https://github.com/murillotadeo/discsocket",
    project_urls={
        "Issue tracker": "https://github.com/murillotadeo/discsocket/issues"
    },
    version="0.1.2",
    packages=['discsocket', 'discsocket.utils', 'discsocket.models'],
    license="MIT",
    description="Python framework for Discord interactions.",
    long_description=readme,
    long_description_content_type="text/markdown",
    include_package_data=True,
    install_requires=['aiohttp>=3.6.0,<3.8.0'],
    python_require=">=3.8.0",
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9'
    ],
    
)